extern crate hercules_ir;

use std::collections::HashMap;
use std::iter::repeat;

use self::hercules_ir::*;

/*
 * Pretty much all of the codegen functions need to take in some large subset of
 * IR structures, analysis results, and global pieces of information. Package
 * them all in this struct, and make all the codegen functions members of this
 * struct to cut down on the number of function arguments. This structure
 * shouldn't be modified after creation.
 */
pub(crate) struct FunctionContext<'a> {
    pub(crate) function: &'a Function,
    pub(crate) types: &'a Vec<Type>,
    pub(crate) constants: &'a Vec<Constant>,
    pub(crate) dynamic_constants: &'a Vec<DynamicConstant>,
    pub(crate) def_use: &'a ImmutableDefUseMap,
    pub(crate) reverse_postorder: &'a Vec<NodeID>,
    pub(crate) typing: &'a Vec<TypeID>,
    pub(crate) control_subgraph: &'a Subgraph,
    pub(crate) fork_join_map: &'a HashMap<NodeID, NodeID>,
    pub(crate) fork_join_nest: &'a HashMap<NodeID, Vec<NodeID>>,
    pub(crate) antideps: &'a Vec<(NodeID, NodeID)>,
    pub(crate) bbs: &'a Vec<NodeID>,
    pub(crate) plan: &'a Plan,
    pub(crate) llvm_types: &'a Vec<String>,
    pub(crate) llvm_constants: &'a Vec<String>,
    pub(crate) llvm_dynamic_constants: &'a Vec<String>,
    pub(crate) partitions_inverted_map: Vec<Vec<NodeID>>,
}

impl<'a> FunctionContext<'a> {
    /*
     * Find data inputs to a partition.
     */
    pub(crate) fn partition_data_inputs(&self, partition_id: PartitionID) -> Vec<NodeID> {
        let partition = &self.partitions_inverted_map[partition_id.idx()];

        let mut data_inputs: Vec<NodeID> = partition
            .iter()
            .map(|id| {
                // For each node in the partition, filter out the uses that are
                // data nodes and are in a different partition.
                get_uses(&self.function.nodes[id.idx()])
                    .as_ref()
                    .into_iter()
                    .filter(|id| {
                        // Filter out control nodes (just looking for data
                        // inputs here), check that it's in another partition,
                        // and ignore parameters, constants, and dynamic
                        // constants (those are each passed to partition
                        // functions using different mechanisms).
                        !self.function.nodes[id.idx()].is_control()
                            && self.plan.partitions[id.idx()] != partition_id
                            && !self.function.nodes[id.idx()].is_parameter()
                            && !self.function.nodes[id.idx()].is_constant()
                            && !self.function.nodes[id.idx()].is_dynamic_constant()
                    })
                    .map(|x| *x)
                    .collect::<Vec<NodeID>>()
            })
            // Collect all such uses across the whole partition.
            .flatten()
            .collect();

        // Inputs and outputs of partitions need to be sorted so datums don't
        // get mixed up.
        data_inputs.sort();
        data_inputs
    }

    /*
     * Find data outputs of a partition.
     */
    pub(crate) fn partition_data_outputs(&self, partition_id: PartitionID) -> Vec<NodeID> {
        let partition = &self.partitions_inverted_map[partition_id.idx()];

        let mut data_outputs: Vec<NodeID> = partition
            .iter()
            .filter(|id| {
                // For each data node in the partition, check if it has any uses
                // outside its partition. Users can be control or data nodes.
                // Also, don't add parameter, constant, and dynamic constant
                // nodes. These nodes are passed to partition mechanisms using
                // different mechanism.
                !self.function.nodes[id.idx()].is_control()
                    && !self.function.nodes[id.idx()].is_parameter()
                    && !self.function.nodes[id.idx()].is_constant()
                    && !self.function.nodes[id.idx()].is_dynamic_constant()
                    && self
                        .def_use
                        .get_users(**id)
                        .as_ref()
                        .into_iter()
                        .filter(|id| self.plan.partitions[id.idx()] != partition_id)
                        .map(|x| *x)
                        .count()
                        > 0
            })
            .map(|x| *x)
            // If this partition contains a return node, the data input of that
            // node is a data output.
            .chain(partition.iter().filter_map(|id| {
                if let Node::Return { control: _, data } = self.function.nodes[id.idx()] {
                    Some(data)
                } else {
                    None
                }
            }))
            .collect();

        // Inputs and outputs of partitions need to be sorted so datums don't
        // get mixed up.
        data_outputs.sort();
        data_outputs
    }

    /*
     * Find control nodes that will return from a partition.
     */
    pub(crate) fn partition_control_returns(&self, partition_id: PartitionID) -> Vec<NodeID> {
        let partition = &self.partitions_inverted_map[partition_id.idx()];

        partition
            .iter()
            .filter(|id| {
                // For each control node in the partition, check if it has any
                // users outside its partition. Users can be control nodes - if
                // a user in a different partition is a data node, then the
                // partition is malformed. Return nodes are also unconditionally
                // a control return of this partition.
                let outside_user_count = self
                    .def_use
                    .get_users(**id)
                    .as_ref()
                    .into_iter()
                    .filter(|user_id| {
                        // Users of control nodes can only be data nodes
                        // if they are in the same partition as the
                        // control node. Only control users may be in a
                        // different partition.
                        assert!(
                            !self.function.nodes[id.idx()].is_control()
                                || self.function.nodes[user_id.idx()].is_control()
                                || self.plan.partitions[user_id.idx()] == partition_id
                        );
                        self.plan.partitions[user_id.idx()] != partition_id
                    })
                    .count();

                // Just calculated for the below assert.
                let control_user_count = self
                    .def_use
                    .get_users(**id)
                    .as_ref()
                    .into_iter()
                    .filter(|id| self.function.nodes[id.idx()].is_control())
                    .count();

                // A control node cannot have users inside and outside its own
                // partition. This is because a well-formedness condition of if
                // and match nodes (the only control nodes allowed to have
                // multiple users) is their read successors must be in the same
                // partition as them.
                assert!(
                    !self.function.nodes[id.idx()].is_control()
                        || outside_user_count == 0
                        || outside_user_count == control_user_count
                );
                self.function.nodes[id.idx()].is_control()
                    && (self.function.nodes[id.idx()].is_return() || outside_user_count > 0)
            })
            .map(|x| *x)
            .collect()
    }

    /*
     * Find control successors of a given partition. A partition cannot be a
     * control successor of itself, since a self-cycle is represented as control
     * flow within a partiion. In other words, the graph of control flow between
     * partitions is free of self-loops (an edge connecting a partition to
     * itself).
     */
    pub(crate) fn partition_control_successors(
        &self,
        partition_id: PartitionID,
    ) -> Vec<(PartitionID, NodeID)> {
        let partition = &self.partitions_inverted_map[partition_id.idx()];

        partition
            .iter()
            // Only consider nodes in other partitions that are successors of
            // control nodes. These are necessarily other control nodes.
            .filter(|id| self.function.nodes[id.idx()].is_control())
            .map(|id| {
                // Get the partitions (that are not this partition) of successor
                // nodes of control nodes.
                self.def_use
                    .get_users(*id)
                    .as_ref()
                    .into_iter()
                    .map(|id| self.plan.partitions[id.idx()])
                    .filter(|id| *id != partition_id)
                    .map(move |part_id| (part_id, *id))
            })
            // We want a flat list of all such partitions.
            .flatten()
            .collect()
    }

    /*
     * Calculate the reverse postorder of just this partition.
     */
    pub(crate) fn partition_reverse_postorder(&self, partition_id: PartitionID) -> Vec<NodeID> {
        self.reverse_postorder
            .iter()
            .filter(|id| self.plan.partitions[id.idx()] == partition_id)
            .map(|x| *x)
            .collect()
    }

    /*
     * Determine the array constant inputs to all partition functions. Get the
     * constant IDs, and the array type IDs. Sort by constant ID for
     * consistency.
     */
    pub(crate) fn partition_array_constant_inputs(&self) -> Vec<(ConstantID, TypeID)> {
        let mut res = (0..self.constants.len())
            .filter_map(|idx| {
                self.constants[idx]
                    .try_array_type(self.types)
                    .map(|ty_id| (ConstantID::new(idx), ty_id))
            })
            .collect::<Vec<_>>();

        res.sort();
        res
    }

    /*
     * Determine the dynamic constant inputs to all partition functions. Just
     * assemble the dynamic constant IDs, since the type is always u64. Sort the
     * parameters for consistency.
     */
    pub(crate) fn partition_dynamic_constant_inputs(&self) -> Vec<DynamicConstantID> {
        let mut res = (0..self.dynamic_constants.len())
            .filter_map(|idx| {
                if self.dynamic_constants[idx].is_parameter() {
                    Some(DynamicConstantID::new(idx))
                } else {
                    None
                }
            })
            .collect::<Vec<_>>();

        res.sort();
        res
    }

    /*
     * Determine the array numbers for all the array constants. These are needed
     * to know which pointer passed to the runtime corresponds to which array
     * constant. Return a map from constant ID to array number - non-array
     * constants don't have an array number.
     */
    pub(crate) fn array_constant_inputs(&self) -> Vec<Option<u32>> {
        self.constants
            .iter()
            .scan(0, |num, cons| {
                if cons.try_array_type(self.types).is_some() {
                    let res = Some(*num);
                    *num += 1;
                    Some(res)
                } else {
                    Some(None)
                }
            })
            .collect()
    }
}

/*
 * When emitting individual nodes in the partition codegen functions, a bunch of
 * partition analysis results are needed. Package them all in this struct, and
 * make all of the subroutines of the top level partition codegen functions
 * members of this struct to cut down on the number of function arguments. This
 * structure shouldn't be modified after creation. This structure only holds per
 * partition specific information - for example, global function parameters,
 * constant parameters, and dynamic constant parameters are not stored, since
 * those don't vary across partitions.
 */
pub(crate) struct PartitionContext<'a> {
    pub(crate) function: &'a FunctionContext<'a>,
    pub(crate) partition_id: PartitionID,
    pub(crate) top_node: NodeID,
    pub(crate) data_inputs: Vec<NodeID>,
    pub(crate) data_outputs: Vec<NodeID>,
    pub(crate) control_returns: Vec<NodeID>,
    pub(crate) reverse_postorder: Vec<NodeID>,
    pub(crate) partition_input_types: Vec<TypeID>,
    pub(crate) return_type: Type,
    pub(crate) manifest: PartitionManifest,
}

impl<'a> PartitionContext<'a> {
    pub(crate) fn new(
        function: &'a FunctionContext<'a>,
        partition_id: PartitionID,
        top_node: NodeID,
    ) -> Self {
        let data_inputs = function.partition_data_inputs(partition_id);
        let data_outputs = function.partition_data_outputs(partition_id);
        let control_returns = function.partition_control_returns(partition_id);
        let control_successors = function.partition_control_successors(partition_id);
        let reverse_postorder = function.partition_reverse_postorder(partition_id);

        // The data input types are just the types of data nodes used by this
        // partition, originating in another partition.
        let partition_input_types = data_inputs
            .iter()
            .map(|id| function.typing[id.idx()])
            .collect();

        // The return struct contains all of the data outputs, plus control
        // information if there are multiple successor partitions. The control
        // information is used by the Hercules runtime to implement control flow
        // between partitions.
        let multiple_control_successors = control_successors.len() > 1;
        let output_data_types = data_outputs.iter().map(|id| function.typing[id.idx()]);
        let return_type = if multiple_control_successors {
            let u64_ty_id = TypeID::new(
                function
                    .types
                    .iter()
                    .position(|ty| *ty == Type::UnsignedInteger64)
                    .unwrap(),
            );
            Type::Product(
                output_data_types
                    .chain(std::iter::once(u64_ty_id))
                    .collect(),
            )
        } else {
            Type::Product(output_data_types.collect())
        };

        // Assemble the manifest.
        let mut manifest = PartitionManifest::default();
        manifest.top_node = top_node;

        // The first inputs are the data inputs, from other partitions.
        manifest
            .inputs
            .extend(data_inputs.iter().map(|x| PartitionInput::DataInput(*x)));

        // The next inputs are the function parameters, all in order.
        manifest.inputs.extend(
            (0..function.function.param_types.len())
                .map(|x| PartitionInput::FunctionArgument(x as u32)),
        );

        // The next inputs are the array constants, all in order. TODO: only
        // include constant inputs for constants actually used by this function,
        // not all the constants in the module.
        manifest.inputs.extend(
            (0..(function
                .constants
                .iter()
                .filter(|cons| cons.try_array_type(function.types).is_some())
                .count()))
                .map(|x| PartitionInput::ArrayConstant(x as u32)),
        );

        // The last inputs are the dynamic constants, all in order.
        manifest.inputs.extend(
            (0..function.function.num_dynamic_constants)
                .map(|x| PartitionInput::DynamicConstant(x as u32)),
        );

        // The outputs are the data outputs of this partition.
        manifest
            .outputs
            .extend(data_outputs.iter().map(|x| PartitionOutput::DataOutput(*x)));

        // If there are multiple control returns, also output the node being
        // returned from.
        if multiple_control_successors {
            manifest.outputs.push(PartitionOutput::ControlIndicator);
        }

        // Store the successor partitions.
        manifest.successor_partitions = control_successors
            .into_iter()
            .map(|(part_id, control_id)| (control_id, part_id))
            .collect();

        PartitionContext {
            function,
            partition_id,
            top_node,
            data_inputs,
            data_outputs,
            control_returns,
            reverse_postorder,
            partition_input_types,
            return_type,
            manifest,
        }
    }
}

/*
 * Types, constants, and dynamic constants are fairly simple to translate into
 * LLVM IR.
 */

pub(crate) fn generate_type_string(ty: &Type, llvm_types: &Vec<String>) -> String {
    match ty {
        Type::Control(_) => {
            // Later, we create virtual registers corresponding to fork nodes of
            // type i64, so we need the "type" of the fork node to be i64.
            "i64".to_string()
        }
        Type::Boolean => "i1".to_string(),
        Type::Integer8 | Type::UnsignedInteger8 => "i8".to_string(),
        Type::Integer16 | Type::UnsignedInteger16 => "i16".to_string(),
        Type::Integer32 | Type::UnsignedInteger32 => "i32".to_string(),
        Type::Integer64 | Type::UnsignedInteger64 => "i64".to_string(),
        Type::Float32 => "float".to_string(),
        Type::Float64 => "double".to_string(),
        // Because we traverse in bottom-up order, we can assume that the LLVM
        // types for children types are already computed.
        Type::Product(fields) => {
            let mut iter = fields.iter();
            if let Some(first) = iter.next() {
                iter.fold("{".to_string() + &llvm_types[first.idx()], |s, f| {
                    s + ", " + &llvm_types[f.idx()]
                }) + "}"
            } else {
                "{}".to_string()
            }
        }
        Type::Summation(_) => todo!(),
        Type::Array(_, _) => {
            // Array types becomes pointers. The element type and dynamic
            // constant bounds characterize the access code we generate later,
            // not the type itself.
            "ptr".to_string()
        }
    }
}

pub(crate) fn generate_type_strings(module: &Module) -> Vec<String> {
    // Render types into LLVM IR. This requires translating from our interning
    // structures to LLVM types. We can't just blow through the types vector,
    // since a type may reference a type ID ahead of it in the vector. Instead,
    // iterate types in a bottom up order with respect to the type intern DAGs.
    let mut llvm_types = vec!["".to_string(); module.types.len()];
    for id in types_bottom_up(&module.types) {
        llvm_types[id.idx()] = generate_type_string(&module.types[id.idx()], &llvm_types);
    }

    llvm_types
}

pub(crate) fn generate_constant_string(
    cons_id: ConstantID,
    cons: &Constant,
    tys: &Vec<Type>,
    llvm_constants: &Vec<String>,
) -> String {
    match cons {
        Constant::Boolean(val) => {
            if *val {
                "true".to_string()
            } else {
                "false".to_string()
            }
        }
        Constant::Integer8(val) => format!("{}", val),
        Constant::Integer16(val) => format!("{}", val),
        Constant::Integer32(val) => format!("{}", val),
        Constant::Integer64(val) => format!("{}", val),
        Constant::UnsignedInteger8(val) => format!("{}", val),
        Constant::UnsignedInteger16(val) => format!("{}", val),
        Constant::UnsignedInteger32(val) => format!("{}", val),
        Constant::UnsignedInteger64(val) => format!("{}", val),
        Constant::Float32(val) => {
            if val.fract() == 0.0 {
                format!("{}.0", val)
            } else {
                format!("{}", val)
            }
        }
        Constant::Float64(val) => {
            if val.fract() == 0.0 {
                format!("{}.0", val)
            } else {
                format!("{}", val)
            }
        }
        Constant::Product(_, _) | Constant::Summation(_, _, _) | Constant::Array(_, _) => {
            format!("%cons.{}", cons_id.idx())
        }
        Constant::Zero(ty_id) => match tys[ty_id.idx()] {
            Type::Product(_) | Type::Summation(_) | Type::Array(_, _) => {
                format!("%cons.{}", cons_id.idx())
            }
            _ => "zeroinitializer".to_string(),
        },
    }
}

pub(crate) fn generate_constant_strings(module: &Module) -> Vec<String> {
    // Render constants into LLVM IR. This is done in a very similar manner as
    // types.
    let mut llvm_constants = vec!["".to_string(); module.constants.len()];
    for id in constants_bottom_up(&module.constants) {
        llvm_constants[id.idx()] = generate_constant_string(
            id,
            &module.constants[id.idx()],
            &module.types,
            &llvm_constants,
        );
    }

    llvm_constants
}

pub(crate) fn generate_dynamic_constant_strings(module: &Module) -> Vec<String> {
    // Render dynamic constants into LLVM IR.
    let mut llvm_dynamic_constants = vec!["".to_string(); module.dynamic_constants.len()];
    for id in (0..module.dynamic_constants.len()).map(DynamicConstantID::new) {
        match &module.dynamic_constants[id.idx()] {
            DynamicConstant::Constant(val) => llvm_dynamic_constants[id.idx()] = format!("{}", val),
            DynamicConstant::Parameter(_) => {
                llvm_dynamic_constants[id.idx()] = format!("%dyn_cons.{}", id.idx())
            }
        }
    }

    llvm_dynamic_constants
}

/*
 * Calculate in-memory size and alignment of a type. The size is optional, since
 * array types with dynamic constant dimensions may not have a compile time
 * known size.
 */
pub fn type_size_and_alignment(module: &Module, ty: TypeID) -> (Option<usize>, usize) {
    match module.types[ty.idx()] {
        Type::Control(_) => {
            panic!("PANIC: Can't calculate in-memory size and alignment of control type.")
        }
        Type::Boolean => (Some(1), 1),
        Type::Integer8 => (Some(1), 1),
        Type::Integer16 => (Some(2), 2),
        Type::Integer32 => (Some(4), 4),
        Type::Integer64 => (Some(8), 8),
        Type::UnsignedInteger8 => (Some(1), 1),
        Type::UnsignedInteger16 => (Some(2), 2),
        Type::UnsignedInteger32 => (Some(4), 4),
        Type::UnsignedInteger64 => (Some(8), 8),
        Type::Float32 => (Some(4), 4),
        Type::Float64 => (Some(8), 8),
        Type::Product(ref fields) => {
            let (size, align) = fields
                .iter()
                .map(|ty| type_size_and_alignment(module, *ty))
                .fold(
                    (Some(0), 1),
                    |(acc_size, acc_align), (field_size, field_align)| {
                        // Alignment of product is maximum alignment of fields.
                        let new_align = std::cmp::max(acc_align, field_align);
                        if let (Some(acc_size), Some(field_size)) = (acc_size, field_size) {
                            // Pre-padding is so that the new field has proper
                            // alignment within the product.
                            let mut pre_padding = field_align - acc_size % field_align;
                            if pre_padding == field_align {
                                pre_padding = 0;
                            }
                            (Some(acc_size + pre_padding + field_size), new_align)
                        } else {
                            (None, new_align)
                        }
                    },
                );

            if let Some(size) = size {
                // Post-padding is so that the overall in-memory size has the
                // right alignment in an array, and is only done at the end.
                let mut post_padding = align - size % align;
                if post_padding == align {
                    post_padding = 0;
                }
                (Some(size + post_padding), align)
            } else {
                (None, align)
            }
        }
        Type::Summation(_) => todo!(),
        Type::Array(elem, ref dims) => {
            let (maybe_elem_size, elem_align) = type_size_and_alignment(module, elem);

            // We can only calculate the number of elements at compile time if
            // every dynamic constant dimension is a compile-time constant.
            let maybe_num_elems = dims.iter().fold(Some(1), |acc, dim| {
                Some(acc? * module.dynamic_constants[dim.idx()].try_constant()?)
            });

            // Even if the number of elements is compile-time known, the element
            // type may have unknown compile-time size.
            if let (Some(elem_size), Some(num_elems)) = (maybe_elem_size, maybe_num_elems) {
                (Some(elem_size * num_elems), elem_align)
            } else {
                (None, elem_align)
            }
        }
    }
}

/*
 * Calculate in-memory bytes representing constant. Return the in-memory bytes
 * and the alignment of the constant, if it's non-zero. If it's zero, optionally
 * return the size of the constant, and its alignment. TODO: decide on how to
 * represent memory layouts at the compiler level.
 */
pub fn embed_constant(module: &Module, cons: ConstantID) -> ConstantBytes {
    let unchecked = match module.constants[cons.idx()] {
        // Handle zero constant scalars below.
        Constant::Boolean(v) => ConstantBytes::NonZero(vec![v as u8], 1),
        Constant::Integer8(v) => ConstantBytes::NonZero(v.to_ne_bytes().to_vec(), 1),
        Constant::Integer16(v) => ConstantBytes::NonZero(v.to_ne_bytes().to_vec(), 2),
        Constant::Integer32(v) => ConstantBytes::NonZero(v.to_ne_bytes().to_vec(), 4),
        Constant::Integer64(v) => ConstantBytes::NonZero(v.to_ne_bytes().to_vec(), 8),
        Constant::UnsignedInteger8(v) => ConstantBytes::NonZero(v.to_ne_bytes().to_vec(), 1),
        Constant::UnsignedInteger16(v) => ConstantBytes::NonZero(v.to_ne_bytes().to_vec(), 2),
        Constant::UnsignedInteger32(v) => ConstantBytes::NonZero(v.to_ne_bytes().to_vec(), 4),
        Constant::UnsignedInteger64(v) => ConstantBytes::NonZero(v.to_ne_bytes().to_vec(), 8),
        Constant::Float32(v) => ConstantBytes::NonZero(v.to_ne_bytes().to_vec(), 4),
        Constant::Float64(v) => ConstantBytes::NonZero(v.to_ne_bytes().to_vec(), 8),
        Constant::Product(ty, ref fields) => {
            let field_bytes: Vec<ConstantBytes> = fields
                .iter()
                .map(|field| embed_constant(module, *field))
                .collect();

            if field_bytes.iter().all(|cb| {
                if let ConstantBytes::Zero(_, _) = cb {
                    true
                } else {
                    false
                }
            }) {
                // If all of the fields are zero constants, then this is a zero
                // constant.
                let (size, align) = type_size_and_alignment(module, ty);
                ConstantBytes::Zero(size, align)
            } else {
                // We only construct the in-memory bytes if there is a non-zero
                // bytes somewhere.
                let (mut bytes, align) = field_bytes.into_iter().fold(
                    (vec![], 0),
                    |(mut acc_bytes, acc_align), field| {
                        // Alignment of product is maximum alignment of fields.
                        let new_align = std::cmp::max(acc_align, field.align());

                        // Pre-padding is so that the new field has proper
                        // alignment within the product.
                        while acc_bytes.len() % field.align() != 0 {
                            acc_bytes.push(0);
                        }
                        match field {
                            ConstantBytes::NonZero(bytes, _) => acc_bytes.extend(&bytes),
                            ConstantBytes::Zero(size, _) => acc_bytes.extend(repeat(0).take(size.expect("PANIC: Attempted to embed a zero constant with unknown size into a non-zero constant product. Non-zero constants must have compile-time known size. This is probably because an array field is a zero constant with non-constant dynamic constant dimensions."))),
                        }
                        (acc_bytes, new_align)
                    },
                );

                // Post-padding is so that the overall in-memory vector has the
                // right size, and is only done at the end.
                while bytes.len() % align != 0 {
                    bytes.push(0);
                }
                ConstantBytes::NonZero(bytes, align)
            }
        }
        Constant::Summation(_, _, _) => todo!(),
        Constant::Array(ty, ref elements) => {
            let element_bytes: Vec<ConstantBytes> = elements
                .iter()
                .map(|element| embed_constant(module, *element))
                .collect();

            let (size, align) = type_size_and_alignment(module, ty);
            if element_bytes.iter().all(|cb| {
                if let ConstantBytes::Zero(_, _) = cb {
                    true
                } else {
                    false
                }
            }) {
                // If all of the fields are zero constants, then this is a zero
                // constant.
                ConstantBytes::Zero(size, align)
            } else {
                let array_bytes: Vec<u8> = element_bytes
                    .into_iter()
                    .map(|cb| match cb {
                        ConstantBytes::NonZero(bytes, _) => bytes,
                        ConstantBytes::Zero(size, _) => vec![0; size.expect("PANIC: Attempted to embed a zero constant with unknown size into a non-zero constant array. Non-zero constants must have compile-time known size. This is probably because an array element is a zero constant with non-constant dynamic constant dimensions.")],
                    })
                    .flatten()
                    .collect();
                assert_eq!(array_bytes.len(), size.expect("PANIC: Size of a non-zero constant array is unknown at compile time. All non-zero constants must have compile time known size."), "PANIC: Size of array type calculated by type_size_and_alignment differs from calculated in-memory byte representation's size.");
                ConstantBytes::NonZero(array_bytes, align)
            }
        }
        Constant::Zero(ty) => {
            let (size, align) = type_size_and_alignment(module, ty);
            ConstantBytes::Zero(size, align)
        }
    };

    // Catch all code for making zero constant scalars actually
    // ConstantBytes::Zero variants.
    if let ConstantBytes::NonZero(bytes, align) = &unchecked {
        if module.constants[cons.idx()].is_strictly_scalar() && bytes.iter().all(|x| *x == 0) {
            return ConstantBytes::Zero(Some(bytes.len()), *align);
        }
    }
    unchecked
}
