extern crate hercules_ir;

use std::collections::HashMap;
use std::fmt::Write;

use self::hercules_ir::*;

use crate::*;

/*
 * Top level function to generate code for a module. Emits LLVM IR text. Calls
 * out to backends to generate code for individual partitions. Creates a
 * manifest describing the generated code.
 */
pub fn codegen<W: Write>(
    module: &Module,
    def_uses: &Vec<ImmutableDefUseMap>,
    reverse_postorders: &Vec<Vec<NodeID>>,
    typing: &ModuleTyping,
    control_subgraphs: &Vec<Subgraph>,
    fork_join_maps: &Vec<HashMap<NodeID, NodeID>>,
    fork_join_nests: &Vec<HashMap<NodeID, Vec<NodeID>>>,
    antideps: &Vec<Vec<(NodeID, NodeID)>>,
    bbs: &Vec<Vec<NodeID>>,
    plans: &Vec<Plan>,
    w: &mut W,
) -> Result<ModuleManifest, std::fmt::Error> {
    // Render types, constants, and dynamic constants into LLVM IR.
    let llvm_types = generate_type_strings(module);
    let llvm_constants = generate_constant_strings(module);
    let llvm_dynamic_constants = generate_dynamic_constant_strings(module);

    // Generate a dummy uninitialized global - this is needed so that there'll
    // be a non-empty .bss section in the ELF object file.
    write!(w, "@dummy = dso_local global i32 0, align 4\n")?;

    // Do codegen for each function individually. Get each function's manifest.
    let mut manifests = vec![];
    for function_idx in 0..module.functions.len() {
        // There's a bunch of per-function information we use.
        let context = FunctionContext {
            function: &module.functions[function_idx],
            types: &module.types,
            constants: &module.constants,
            dynamic_constants: &module.dynamic_constants,
            def_use: &def_uses[function_idx],
            reverse_postorder: &reverse_postorders[function_idx],
            typing: &typing[function_idx],
            control_subgraph: &control_subgraphs[function_idx],
            fork_join_map: &fork_join_maps[function_idx],
            fork_join_nest: &fork_join_nests[function_idx],
            antideps: &antideps[function_idx],
            bbs: &bbs[function_idx],
            plan: &plans[function_idx],
            llvm_types: &llvm_types,
            llvm_constants: &llvm_constants,
            llvm_dynamic_constants: &llvm_dynamic_constants,
            partitions_inverted_map: plans[function_idx].invert_partition_map(),
        };

        manifests.push(context.codegen_function(w)?);
    }

    // Assemble the manifest for the whole module.
    Ok(ModuleManifest {
        functions: manifests,
        types: module.types.clone(),
        type_sizes_aligns: (0..module.types.len())
            .map(|idx| {
                if module.types[idx].is_control() {
                    (None, 0)
                } else {
                    type_size_and_alignment(module, TypeID::new(idx))
                }
            })
            .collect(),
        dynamic_constants: module.dynamic_constants.clone(),
        // Get the types of all of the constants. This requires collecting over
        // all of the functions, since the calculated types of constants may be
        // distributed over many functions. This may contain duplicate mappings,
        // but this should be fine for our purposes, since the mappings
        // shouldn't conflict.
        constant_types: module
            .functions
            .iter()
            .enumerate()
            .map(|(func_idx, function)| {
                function
                    .nodes
                    .iter()
                    .enumerate()
                    .filter_map(move |(idx, node)| {
                        Some((node.try_constant()?, typing[func_idx][idx]))
                    })
            })
            .flatten()
            .collect(),
        array_constants: (0..module.constants.len())
            .map(ConstantID::new)
            .filter_map(|cons_id| {
                if module.constants[cons_id.idx()]
                    .try_array_type(&module.types)
                    .is_some()
                {
                    Some(embed_constant(module, cons_id))
                } else {
                    None
                }
            })
            .collect(),
        array_cons_ids: (0..module.constants.len())
            .map(ConstantID::new)
            .filter(|id| {
                module.constants[id.idx()]
                    .try_array_type(&module.types)
                    .is_some()
            })
            .collect(),
    })
}

impl<'a> FunctionContext<'a> {
    /*
     * Each function gets codegened separately.
     */
    fn codegen_function<W: Write>(&self, w: &mut W) -> Result<FunctionManifest, std::fmt::Error> {
        // Find the "top" control node of each partition. One well-formedness
        // condition of partitions is that there is exactly one "top" control
        // node.
        let top_nodes: Vec<NodeID> = self
            .partitions_inverted_map
            .iter()
            .enumerate()
            .map(|(part_idx, part)| {
                // For each partition, find the "top" node.
                *part
                    .iter()
                    .filter(move |id| {
                        // The "top" node is a control node having at least one
                        // control predecessor in another partition, or is a
                        // start node. Every predecessor in the control subgraph
                        // is a control node.
                        self.function.nodes[id.idx()].is_start()
                            || (self.function.nodes[id.idx()].is_control()
                                && self
                                    .control_subgraph
                                    .preds(**id)
                                    .filter(|pred_id| {
                                        self.plan.partitions[pred_id.idx()].idx() != part_idx
                                    })
                                    .count()
                                    > 0)
                    })
                    .next()
                    .unwrap()
            })
            .collect();

        // Collect all the node IDs that are values returned by this function.
        let returned_values = self
            .function
            .nodes
            .iter()
            .filter_map(|node| node.try_return().map(|(_, data)| data))
            .collect();

        // Get the partition ID of the start node.
        let top_partition = self.plan.partitions[0];

        // Generate code for each individual partition. This generates a single
        // LLVM function per partition. These functions will be called in async
        // tasks by the Hercules runtime.
        assert_eq!(self.plan.num_partitions, top_nodes.len());
        let mut manifests = vec![];
        for part_idx in 0..self.plan.num_partitions {
            match self.plan.partition_devices[part_idx] {
                Device::CPU => manifests.push(self.codegen_cpu_partition(top_nodes[part_idx], w)?),
                Device::GPU => todo!(),
            }
        }

        // Assemble the manifest for the whole function.
        Ok(FunctionManifest {
            name: self.function.name.clone(),
            param_types: self.function.param_types.clone(),
            return_type: self.function.return_type,
            typing: self.typing.clone(),
            num_dynamic_constant_parameters: self.function.num_dynamic_constants,
            partitions: manifests,
            // TODO: populate dynamic constant rules.
            dynamic_constant_rules: vec![],
            top_partition,
            returned_values,
        })
    }
}
