extern crate serde;

use self::serde::Deserialize;
use self::serde::Serialize;

use crate::*;

/*
 * Every .hbin file contains a manifest which describes the Hercules functions
 * contained in the module. This information is used by the runtime to execute
 * the functions properly, the chief concern being how to stitch together the
 * execution of each partition.
 */
#[derive(Debug, Serialize, Deserialize, Hash)]
pub struct ModuleManifest {
    // A module contains a manifest per individual function.
    pub functions: Vec<FunctionManifest>,
    // All of the types used in the module.
    pub types: Vec<Type>,
    // All of the sizes and alignments for types in the module.
    pub type_sizes_aligns: Vec<(Option<usize>, usize)>,
    // All of the dynamic constants used in the module.
    pub dynamic_constants: Vec<DynamicConstant>,
    // Store the types of constants.
    pub constant_types: Vec<(ConstantID, TypeID)>,
    // The only constants that aren't baked into the generated code are array
    // constants. These are explicitly stored in and loaded from the manifest.
    // We only store the array bytes if the array is non-zero. If the array is
    // all zeros, we optionally store the size of the array. We may not be able
    // to store the array size at compile time if it's all zeros, as its size
    // may be a dynamic constant with compile-time unknown size.
    pub array_constants: Vec<ConstantBytes>,
    // Store map from array number, in array_constants, to constant ID.
    pub array_cons_ids: Vec<ConstantID>,
}

/*
 * embed_constant calculates the byte representation of a constant. For zero
 * constants, we avoid storing the actual zero bytes, and optionally store the
 * size - zero constant arrays may have dynamic constant dimensions unknown at
 * compile time.
 */
#[derive(Debug, Clone, PartialEq, Eq, Hash, Serialize, Deserialize)]
pub enum ConstantBytes {
    NonZero(Vec<u8>, usize),
    Zero(Option<usize>, usize),
}

#[derive(Debug, Serialize, Deserialize, Hash)]
pub struct FunctionManifest {
    pub name: String,
    // Types of the function parameters.
    pub param_types: Vec<TypeID>,
    // Type of the return value.
    pub return_type: TypeID,
    // Types of all of the nodes in this function. Used for figuring out the
    // type of partition data inputs and outputs.
    pub typing: Vec<TypeID>,
    // Number of dynamic constant parameters that need to provided.
    pub num_dynamic_constant_parameters: u32,
    // Manifests for constituent partitions.
    pub partitions: Vec<PartitionManifest>,
    // When using dynamic constants, certain constraints are generated. For
    // example, if one uses division in dynamic constant math, the resulting
    // dynamic constant must be an integer, so the numerator dynamic constant
    // must be divisible by the denominator dynamic constant. These are stored
    // per function, since different functions have different contraints on
    // their dynamic constant parameters.
    pub dynamic_constant_rules: Vec<DynamicConstantRule>,
    // The partition containing the start node might not be partition 0.
    pub top_partition: PartitionID,
    // Keep track of which node values are returned from the overall function.
    pub returned_values: Vec<NodeID>,
}

/*
 * Rules for validity of provided dynamic constants
 */
#[derive(Debug, Serialize, Deserialize, Hash, PartialEq, Eq)]
pub enum DynamicConstantRule {
    // Generated from subtraction.
    LessThan(DynamicConstantID, DynamicConstantID),
    // Generated from division.
    Divides(DynamicConstantID, DynamicConstantID),
}

#[derive(Debug, Serialize, Deserialize, Hash, PartialEq, Eq)]
pub enum PartitionInput {
    // Data input from another partition within this function.
    DataInput(NodeID),
    // An argument from the function parameters. Integer is the parameter index.
    FunctionArgument(u32),
    // An array constant used in this function. Integer is the array constant
    // number.
    ArrayConstant(u32),
    // A dynamic constant parameter of this function. Integer is the dynamic
    // constant parameter number.
    DynamicConstant(u32),
}

#[derive(Debug, Serialize, Deserialize, Hash, PartialEq, Eq)]
pub enum PartitionOutput {
    // Data output used by another partition within this function, or to be
    // returned from this function.
    DataOutput(NodeID),
    // Value indicating control flow that the runtime should take.
    ControlIndicator,
}

#[derive(Debug, Serialize, Deserialize, Default, Hash)]
pub struct PartitionManifest {
    // Top node for this partition.
    pub top_node: NodeID,
    pub inputs: Vec<PartitionInput>,
    pub outputs: Vec<PartitionOutput>,

    // Store the partitions that come after this partition. The first element of
    // the pair is the control returning node in the current partition that
    // corresponds to a particular successor, and the second element of the pair
    // is the partition ID of the corresponding successor partition.
    pub successor_partitions: Vec<(NodeID, PartitionID)>,
}

impl ConstantBytes {
    pub fn align(&self) -> usize {
        match self {
            ConstantBytes::NonZero(_, align) => *align,
            ConstantBytes::Zero(_, align) => *align,
        }
    }
}

/*
 * Get the data inputs of partitions in a function.
 */
pub fn function_partition_inputs(manifest: &FunctionManifest) -> Vec<NodeID> {
    let mut all_data_inputs = manifest
        .partitions
        .iter()
        .map(|manifest| {
            manifest.inputs.iter().filter_map(|input| {
                if let PartitionInput::DataInput(idx) = input {
                    Some(*idx)
                } else {
                    None
                }
            })
        })
        .flatten()
        .collect::<Vec<NodeID>>();
    all_data_inputs.sort();
    all_data_inputs.dedup();
    all_data_inputs
}

/*
 * Get the data outputs of partitions in a function.
 */
pub fn function_partition_outputs(manifest: &FunctionManifest) -> Vec<NodeID> {
    let mut all_data_outputs = manifest
        .partitions
        .iter()
        .map(|manifest| {
            manifest.outputs.iter().filter_map(|output| {
                if let PartitionOutput::DataOutput(idx) = output {
                    Some(*idx)
                } else {
                    None
                }
            })
        })
        .flatten()
        .collect::<Vec<NodeID>>();
    all_data_outputs.sort();
    all_data_outputs.dedup();
    all_data_outputs
}

/*
 * Verify a manifest is valid.
 */
pub fn verify_manifest(manifest: &ModuleManifest) -> Result<(), String> {
    manifest
        .functions
        .iter()
        .map(|manifest| {
            // First, check that all partitiion data inputs get used, and that
            // all partition data outputs are produced.
            let mut all_data_inputs = function_partition_inputs(manifest);
            let all_data_outputs = function_partition_outputs(manifest);

            // Data outputs include values returned from the overall function.
            // These are obviously not used by another partition, but that's
            // fine. We add these artificially as "inputs" to a partition so
            // that this check still passes. We don't just remove the returned
            // value from the data outputs because it might also be the
            // legitimate output of another partition, and doing that would
            // cause a false negative.
            all_data_inputs.extend(&manifest.returned_values);
            all_data_inputs.sort();
            all_data_inputs.dedup();
            if all_data_inputs != all_data_outputs {
                return Err(format!("PANIC: Partitions in manifest contain inconsistent data inputs and data outputs of partitions.\nHere are the data input IDs:\n{:?}\nHere are the data output IDs:\n{:?}\n", all_data_inputs, all_data_outputs));
            }

            // Next, check that partitions contain multiple successors if and
            // only if a control indicator is present in their outputs.
            // TODO: this will fail to verify a partition with a return node and
            // a single successor partition, which technically is allowed. In
            // this scenario, a control indicator is required (to tell if we
            // should return or advance the control token), but there is only
            // one successor partition. Deal with this later.
            manifest.partitions.iter().map(|manifest| {
                let multiple_successors = manifest.successor_partitions.len() > 1;
                let control_indicator = manifest.outputs.contains(&PartitionOutput::ControlIndicator);
                if !multiple_successors && control_indicator {
                    Err("PANIC: Partition in manifest doesn't have multiple successor partitions, but has a control indicator output.".to_string())
                } else if multiple_successors && !control_indicator {
                    Err("PANIC: Partition in manifest has multiple successor partitions, but not a control indicator output.".to_string())
                } else {
                    Ok(())
                }
            }).collect::<Result<Vec<()>, String>>()?;

            Ok(())
        })
        .collect::<Result<Vec<()>, String>>()
        .map(|_| ())
}
