#![feature(iter_intersperse)]

extern crate anyhow;
extern crate hercules_ir;
extern crate postcard;
extern crate proc_macro;

use std::ffi::OsStr;
use std::fmt::Write;
use std::fs::File;
use std::hash::{DefaultHasher, Hash, Hasher};
use std::io::prelude::*;
use std::path::Path;

use proc_macro::*;

use self::hercules_ir::*;

/*
 * Parse manifest from header of .hbin file.
 */
fn manifest_and_module_bytes(buffer: &[u8]) -> (ModuleManifest, Vec<u8>) {
    postcard::from_bytes(buffer).unwrap()
}

/*
 * Convert Hercules types to the Rust types generated in the interface.
 */
fn generate_type_string(ty: &Type, rust_types: &Vec<String>) -> String {
    match ty {
        Type::Control(_) => "NOT_A_REAL_TYPE".to_string(),
        Type::Boolean => "bool".to_string(),
        Type::Integer8 => "i8".to_string(),
        Type::Integer16 => "i16".to_string(),
        Type::Integer32 => "i32".to_string(),
        Type::Integer64 => "i64".to_string(),
        Type::UnsignedInteger8 => "u8".to_string(),
        Type::UnsignedInteger16 => "u16".to_string(),
        Type::UnsignedInteger32 => "u32".to_string(),
        Type::UnsignedInteger64 => "u64".to_string(),
        Type::Float32 => "f32".to_string(),
        Type::Float64 => "f64".to_string(),
        Type::Product(fields) => {
            fields
                .iter()
                .map(|field_id| &rust_types[field_id.idx()] as &str)
                .fold("(".to_string(), |acc, field| acc + field + ",")
                + ")"
        }
        Type::Summation(_) => todo!(),
        Type::Array(elem, _) => format!("*mut {}", &rust_types[elem.idx()]),
    }
}

/*
 * Emit a dynamic constant, which isn't necessarily just a dynamic constant
 * parameter.
 */
fn emit_dynamic_constant(dc: DynamicConstantID, manifest: &ModuleManifest) -> String {
    match manifest.dynamic_constants[dc.idx()] {
        DynamicConstant::Constant(val) => format!("{}", val),
        DynamicConstant::Parameter(idx) => format!("dyn_cons_{}", idx),
    }
}

/*
 * Emit dynamic constant math to calculate the runtime size of a type, in number
 * of bytes.
 */
fn emit_type_size_expression(ty: TypeID, manifest: &ModuleManifest) -> String {
    match manifest.types[ty.idx()] {
        Type::Array(elem, ref dcs) => {
            dcs.iter().fold(
                "(".to_string() + &emit_type_size_expression(elem, manifest),
                |acc, dc| acc + " * " + &emit_dynamic_constant(*dc, manifest),
            ) + ")"
        }
        _ => format!(
            "{}",
            manifest.type_sizes_aligns[ty.idx()]
                .0
                .expect("PANIC: Size of non-array type is unknown at compile-time.")
        ),
    }
}

/*
 * Generate async Rust code orchestrating partition execution.
 */
fn codegen(manifest: &ModuleManifest, elf: &[u8]) -> Result<String, anyhow::Error> {
    // Verify the manifest.
    verify_manifest(manifest).map_err(anyhow::Error::msg)?;

    // Convert Hercules IR types to Rust types.
    let mut rust_types = vec!["".to_string(); manifest.types.len()];
    for id in types_bottom_up(&manifest.types) {
        rust_types[id.idx()] = generate_type_string(&manifest.types[id.idx()], &rust_types);
    }

    // Write to a String containing all of the Rust code.
    let mut rust_code = "".to_string();

    // Emit the ELF bytes as a static byte string constant.
    write!(
        rust_code,
        "const __HERCULES_ELF_OBJ: &[u8] = {};",
        Literal::byte_string(elf)
    )?;

    // Emit the async functions.
    for function in &manifest.functions {
        // Emit the function signature.
        write!(rust_code, "async fn {}(", function.name)?;
        let param_tys = function.param_types.iter().map(|id| &rust_types[id.idx()]);
        for (idx, param) in param_tys.enumerate() {
            write!(rust_code, "param_{}: {},", idx, param)?;
        }
        for dyn_cons_idx in 0..function.num_dynamic_constant_parameters {
            write!(rust_code, "dyn_cons_{}: u64,", dyn_cons_idx)?;
        }
        write!(
            rust_code,
            ") -> {} {{",
            rust_types[function.return_type.idx()]
        )?;

        // Check that the ELF got embedded properly.
        write!(
            rust_code,
            "use std::hash::{{DefaultHasher, Hash, Hasher}};debug_assert_eq!({{let mut s = DefaultHasher::new(); __HERCULES_ELF_OBJ.hash(&mut s); s.finish()}}, {});",
            {
                let mut s = DefaultHasher::new();
                elf.hash(&mut s);
                s.finish()
            }
        )?;

        // Load the ELF object, and cast the appropriate pointers.
        write!(
            rust_code,
            r#"let module = ::hercules_rt::Module::new(__HERCULES_ELF_OBJ);"#
        )?;
        for (idx, partition) in function.partitions.iter().enumerate() {
            // Determine the signature for the partition function.
            let u64_ty_id = TypeID::new(
                manifest
                    .types
                    .iter()
                    .position(|ty| *ty == Type::UnsignedInteger64)
                    .unwrap(),
            );
            let input_types = partition.inputs.iter().map(|input| match input {
                PartitionInput::DataInput(node_id) => function.typing[node_id.idx()],
                PartitionInput::FunctionArgument(param_idx) => {
                    function.param_types[*param_idx as usize]
                }
                PartitionInput::ArrayConstant(num) => {
                    // We store the types of constants in the manifest - use
                    // this info to get the type of this array constant.
                    let arr_cons_id = manifest.array_cons_ids[*num as usize];
                    manifest
                        .constant_types
                        .iter()
                        .filter(|(cons_id, _)| arr_cons_id == *cons_id)
                        .next()
                        .expect("PANIC: Couldn't find type of array constant in manifest.")
                        .1
                }
                PartitionInput::DynamicConstant(_) => u64_ty_id,
            });
            let output_type = Type::Product(
                partition
                    .outputs
                    .iter()
                    .map(|input| match input {
                        PartitionOutput::DataOutput(node_id) => function.typing[node_id.idx()],
                        PartitionOutput::ControlIndicator => u64_ty_id,
                    })
                    .collect(),
            );

            // Get the pointer for the partition function, and cast it to the
            // correct function pointer type.
            write!(
                rust_code,
                r#"let fn_ptr_part_{}: extern "C" fn({}) -> {} = unsafe {{ ::core::mem::transmute(module.get_function_ptr("{}_part_{}")) }};"#,
                idx,
                input_types
                    .map(|id| &rust_types[id.idx()] as &str)
                    .intersperse(", ")
                    .fold("".to_string(), |acc, field| acc + field),
                // Generate the output product type on the fly. This type is not
                // necessarily present in the typing of the original IR, before
                // code generation.
                generate_type_string(&output_type, &rust_types),
                function.name,
                idx
            )?;
        }

        // Declare all of the input / output intermediate variables. We declare
        // them as Options since we want to catch use before definition bugs. We
        // get the list of variables from partition outputs specifically, since
        // this also includes returned values.
        let all_partition_outputs = function_partition_outputs(function);
        for node in all_partition_outputs {
            write!(
                rust_code,
                "let mut node_{}: Option<{}> = None;",
                node.idx(),
                &rust_types[function.typing[node.idx()].idx()]
            )?;
        }

        // Declare all of the array constant memories. We declare them as Vecs
        // to allocate the memories. We emit multiplications of the dynamic
        // constant dimensions to allocate the whole memory as one contiguous
        // range. TODO: emit only the array constants actually used in this
        // function.
        for (arr_cons_num, arr_cons_id) in manifest.array_cons_ids.iter().enumerate() {
            let arr_ty_id = manifest
                .constant_types
                .iter()
                .filter(|(cons_id, _)| arr_cons_id == cons_id)
                .next()
                .expect("PANIC: Couldn't find type of array constant in manifest.")
                .1;
            write!(
                rust_code,
                "let mut arr_cons_{}_vec = vec![0u8; {} as usize];let mut arr_cons_{} = arr_cons_{}_vec.as_mut_ptr();arr_cons_{}_vec.leak();",
                arr_cons_num,
                emit_type_size_expression(arr_ty_id, manifest),
                arr_cons_num,
                arr_cons_num,
                arr_cons_num,
            )?;
        }

        // The core executor is a Rust loop. We literally run a "control token"
        // as described in the original sea of nodes paper through the
        // partitions to drive execution.
        write!(
            rust_code,
            "let mut control_token: u32 = {};loop {{",
            function.top_partition.idx()
        )?;

        // Match on the control token position to determine which partition to
        // execute.
        write!(rust_code, "match control_token {{")?;
        for (idx, partition) in function.partitions.iter().enumerate() {
            // The control token stores the current partition index.
            write!(rust_code, "{} => {{", idx)?;

            // Calculate the inputs.
            for (input_idx, input) in partition.inputs.iter().enumerate() {
                write!(
                    rust_code,
                    "let input_{} = {};",
                    input_idx,
                    match input {
                        PartitionInput::DataInput(id) => format!(
                            "node_{}.expect(\"PANIC: Encountered use before def in runtime.\")",
                            id.idx()
                        ),
                        PartitionInput::FunctionArgument(idx) => format!("param_{}", idx),
                        PartitionInput::ArrayConstant(num) => format!("arr_cons_{} as _", num),
                        PartitionInput::DynamicConstant(idx) => format!("dyn_cons_{}", idx),
                    }
                )?;
            }

            // Make the call.
            write!(rust_code, "let output = fn_ptr_part_{}(", idx)?;
            for input_idx in 0..partition.inputs.len() {
                write!(rust_code, "input_{},", input_idx)?;
            }
            write!(rust_code, ");")?;

            // Assign the outputs.
            for (output_idx, output) in partition.outputs.iter().enumerate() {
                write!(
                    rust_code,
                    "{} = Some(output.{});",
                    match output {
                        PartitionOutput::DataOutput(id) => format!("node_{}", id.idx()),
                        // TODO: handle partitions with control indicator
                        // outputs.
                        PartitionOutput::ControlIndicator => "todo!()".to_string(),
                    },
                    output_idx
                )?;
            }

            // If there are no control successors, return from the function. If
            // there is one control successor, unconditionally set the control
            // token to that successor's index. If there are multiple control
            // successors, set the control token to the control indicator output
            // of the partition function.
            // TODO: handle partitions with multiple return nodes.
            // TODO: handle partitions with a return node and a single successor
            // partition.
            if partition.successor_partitions.len() == 0 {
                assert_eq!(function.returned_values.len(), 1);
                write!(
                    rust_code,
                    "return node_{}.expect(\"PANIC: Encountered use before def in runtime.\");",
                    function.returned_values[0].idx()
                )?;
            } else if partition.successor_partitions.len() == 1 {
                write!(
                    rust_code,
                    "control_token = {};",
                    partition.successor_partitions[0].1.idx()
                )?;
            } else {
                todo!();
            }

            // Close the branch.
            write!(rust_code, "}}")?;
        }

        // Close the match, and handle invalid control token values.
        write!(
            rust_code,
            "_ => panic!(\"PANIC: Invalid control token value.\"),}}"
        )?;

        // Close the loop.
        write!(rust_code, "}}")?;

        // Close the function.
        write!(rust_code, "}}")?;
    }

    eprintln!("{}", rust_code);

    Ok(rust_code)
}

/*
 * Load a Hercules compiled module from a .hbin file.
 */
#[proc_macro]
pub fn use_hbin(path: TokenStream) -> TokenStream {
    use TokenTree::Literal;

    // Get the path as a Rust path object, and make sure it's a .hbin file.
    let mut tokens_iter = path.into_iter();
    let token = tokens_iter
        .next()
        .expect("Please provide a path to a .hbin file to the use_hbin! macro.");
    assert!(tokens_iter.next().is_none(), "Too many tokens provided to the use_hbin! macro. Please provide only a path to a .hbin file.");
    let literal = if let Literal(literal) = token {
        literal
    } else {
        panic!("Please provide a string literal containing the path to a .hbin file to the use_hbin! macro.");
    };
    let literal_string = literal.to_string();
    let path = Path::new(&literal_string[1..(literal_string.len() - 1)]);
    assert_eq!(
        path.extension(),
        Some(OsStr::new("hbin")),
        "Please provide only .hbin files to the use_hbin! macro."
    );
    assert_eq!(
        path.try_exists().ok(),
        Some(true),
        "Please provide a valid path to a .hbin file to the use_hbin! macro."
    );

    // Load manifest from path.
    let mut f = File::open(path).unwrap();
    let mut buffer = vec![];
    f.read_to_end(&mut buffer).unwrap();
    let (manifest, elf) = manifest_and_module_bytes(&buffer);
    eprintln!("{:?}", manifest);

    // Generate Rust code.
    codegen(&manifest, &elf).unwrap().parse().unwrap()
}
